-- ***** 3.1 INFO ******
--Quest Log Functions
--* NEW - GetQuestLogSpecialItemCooldown
--* NEW - GetQuestLogSpecialItemInfo
--* NEW - IsQuestLogSpecialItemInRange
--* NEW - UseQuestLogSpecialItem

local _G = getfenv(0);

-- Global Chat Message Function
function AzMsg(msg) DEFAULT_CHAT_FRAME:AddMessage(tostring(msg):gsub("|1","|cffffff80"):gsub("|2","|cffffffff"),0.5,0.75,1.0); end

-- Constants
local modName = "QBar";
local slots = {
	"HeadSlot", "NeckSlot", "ShoulderSlot", "BackSlot", "ChestSlot", "ShirtSlot", "TabardSlot", "WristSlot",
	"HandsSlot", "WaistSlot", "LegsSlot", "FeetSlot", "Finger0Slot", "Finger1Slot", "Trinket0Slot", "Trinket1Slot",
	"MainHandSlot", "SecondaryHandSlot", "RangedSlot",
};

-- Work Vars
local lastName;
local shownItems = 0;
local updateTime = 0;

-- These items are not marked as being quest items, but we want to include them anyway
local qItems = {
	23818,	-- Stillpine Furbolg Language Primer
	23792,	-- Tree Disguise Kit
	24084,	-- Draenei Banner
	24278,	-- Flare Gun
	5456,	-- Divining Scroll (item has no Use: text, even though you can use it)

	11582,	-- Fel Salve
	12922,	-- Empty Canteen
	11914,  -- Cursed Ooze Jar
	11948,	-- Tainted Ooze Jar
	28038,	-- Seaforium PU-36 Explosive Nether Modulator
	28132,	-- Area 52 Special
	23361,	-- Cleansing Vial
	25465,	-- Stormcrow Amulet
	24355,	-- Ironvine Seeds
	25552,	-- Warmaul Ogre Banner
	25555,	-- Kil'sorrow Banner
	25658,	-- Damp Woolen Blanket
	25853,	-- Pack of Incendiary Bombs (Old Hillsbrad)
	24501,	-- Gordawg's Boulder

--	45067,	-- Egg Basket -- Az: offhand item, but I wanted it on my bar for a hotkey
};

-- Config
local cfg;
local defConfig = {
	enabled = true,
	showTips = true,
	vertical = false,
	mirrored = false,
	btnSize = 36,
	scale = 1,
	padding = 1,
};

--------------------------------------------------------------------------------------------------------
--                                                Main                                                --
--------------------------------------------------------------------------------------------------------

local core = CreateFrame("Frame",modName,UIParent);
core:SetMovable(1);
core:SetToplevel(1);
core:SetFrameStrata("MEDIUM");

local function OnMouseUp(self,button)
	core:StopMovingOrSizing();
	cfg.left = core:GetLeft(); cfg.bottom = core:GetBottom();
end

local function OnUpdate(self,elapsed)
	updateTime = (updateTime + elapsed);
	if (updateTime > 0.25) then
		core.UpdateButtons();
		core:SetScript("OnUpdate",nil);
	end
end

core:SetScript("OnMouseDown",core.StartMoving);
core:SetScript("OnMouseUp",OnMouseUp);

core.tip = CreateFrame("GameTooltip",modName.."Tip",nil,"GameTooltipTemplate");
core.tip:SetOwner(UIParent,"ANCHOR_NONE");

core.items = {};

--------------------------------------------------------------------------------------------------------
--                                             Local Funcs                                            --
--------------------------------------------------------------------------------------------------------

local backDrop = { bgFile = "Interface\\ChatFrame\\ChatFrameBackground", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = { left = 3, right = 3, top = 3, bottom = 3 } };

-- Bindings Frame
local function OpenBindingFrame()
	local bindFrame = CreateFrame("Frame",nil,UIParent);
	bindFrame:SetWidth(200);
	bindFrame:SetHeight(60);
	bindFrame:SetPoint("CENTER");
	bindFrame:SetFrameStrata("DIALOG");
	bindFrame:EnableKeyboard(1);
	bindFrame:SetBackdrop(backDrop);
	bindFrame:SetBackdropColor(0.1,0.22,0.35,1);
	bindFrame:SetBackdropBorderColor(0.1,0.1,0.1,1);

	bindFrame:SetScript("OnKeyDown",function(self,key)
		-- Check Key
		if (key == "ESCAPE") then
			if (cfg.bindKey) then
				SetBinding(cfg.bindKey);
				cfg.bindKey = nil;
				core.UpdateBinding();
			end
			AzMsg("|2"..modName.."|r: Binding has been cleared.");
			self:Hide();
			return;
		elseif (key:match("^.SHIFT") or key:match("^.CTRL") or key:match("^.ALT")) then
			return;
		end
		-- Prefix Modifier Key
		if (IsShiftKeyDown()) then
			key = "SHIFT-"..key;
		end
		if (IsControlKeyDown()) then
			key = "CTRL-"..key;
		end
		if (IsAltKeyDown()) then
			key = "ALT-"..key;
		end
		-- Validate
		if (key ~= cfg.bindKey) and (GetBindingAction(key) ~= "") then
			AzMsg("|2Invalid|r: The hotkey |1"..key.."|r is already bound to |1"..GetBindingAction(key).."|r.");
			return;
		end
		-- Clear Old Binding
		if (cfg.bindKey) then
			SetBinding(cfg.bindKey);
		end
		-- Set New Hotkey
		cfg.bindKey = key;
		AzMsg("|2"..modName.."|r: The hotkey has been set to |1"..key.."|r.");
		core.UpdateBinding(1);
		self:Hide();
	end);

	bindFrame.hint = bindFrame:CreateFontString(nil,"ARTWORK","GameFontNormal");
	bindFrame.hint:SetPoint("CENTER");
	bindFrame.hint:SetText("Press the Desired Hotkey...\nEscape to Clear Binding");

	-- Change This Function
	OpenBindingFrame = function() bindFrame:Show(); end
end

-- Lock Frame
local function SetLocked(lock)
	if (lock) then
		core:EnableMouse(nil);
		core:SetBackdrop(nil);
	else
		core:EnableMouse(1);
		core:SetBackdrop(backDrop);
		core:SetBackdropColor(0.1,0.22,0.35,1);
		core:SetBackdropBorderColor(0.1,0.1,0.1,1);
		core:SetWidth(cfg.btnSize + 24);
		core:SetHeight(cfg.btnSize + 24);
	end
end

-- SetEnabled
local function SetEnabledStatus()
	if (cfg.enabled) then
		core:RegisterEvent("BAG_UPDATE");
		core:RegisterEvent("UNIT_INVENTORY_CHANGED");
		core:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN");
		core:RegisterEvent("PLAYER_REGEN_ENABLED");
		core:RegisterEvent("UPDATE_BINDINGS");
--		core:RegisterEvent("PLAYER_ENTERING_WORLD");
		core:RegisterEvent("ZONE_CHANGED_NEW_AREA");	-- Should work better than PLAYER_ENTERING_WORLD
		core:Show();
	else
		core:UnregisterAllEvents();
		core:Hide();
	end
	core.RequestUpdate();
end

--------------------------------------------------------------------------------------------------------
--                                                Items                                               --
--------------------------------------------------------------------------------------------------------

local anchors = { "ANCHOR_BOTTOMLEFT", "ANCHOR_BOTTOMRIGHT", "ANCHOR_RIGHT", "ANCHOR_LEFT" };

-- Hide GTT
local function HideGTT()
	GameTooltip:Hide();
end

-- Button Scripts
local function Button_OnEnter(self)
	if (cfg.showTips) then
		local x, y = (UIParent:GetWidth() / 2), (UIParent:GetHeight() / 2);
		local left, bottom = self:GetLeft(), self:GetBottom();
		local quadrant = (left > x and bottom > y and 1) or (left < x and bottom > y and 2) or (left < x and bottom < y and 3) or (left > x and bottom < y and 4);
		GameTooltip:SetOwner(self,anchors[quadrant]);
		local bag, slot = self:GetAttribute("bag"), self:GetAttribute("slot");
		if (bag) then
			GameTooltip:SetBagItem(bag,slot);
		else
			GameTooltip:SetInventoryItem("player",slot);
		end
	end
end

-- OnClick
local function Button_OnClick(self,button)
	-- Handle Modified Click
	if (HandleModifiedItemClick(self.link)) then
		return;
	-- Ignore
	elseif (IsShiftKeyDown()) then
		if (not cfg.ignoreList) then
			cfg.ignoreList = {};
		end
		cfg.ignoreList[self.name] = true;
		core.RequestUpdate();
	-- Set Hotkey
	elseif (self.name ~= lastName) then
		lastName = self.name;
		core.UpdateBinding(1);
	end
end

--local function Button_OnUpdate(self,elapsed)
--	if (ItemHasRange(self.link)) and (IsItemInRange(self.link,"target")) then
--		self.icon:SetVertexColor(1,1,1);
--	else
--		self.icon:SetVertexColor(0.8,0.1,0.1);
--	end
--end

-- Make Loot Button
local function CreateItemButton()
	local b = CreateFrame("Button",modName..(#core.items + 1),core,"SecureActionButtonTemplate");
	b:SetWidth(cfg.btnSize);
	b:SetHeight(cfg.btnSize);
	b:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square");
	b:RegisterForClicks("LeftButtonUp","RightButtonUp");
--	b:SetScript("OnUpdate",Button_OnUpdate);
	b:SetScript("OnEnter",Button_OnEnter);
	b:SetScript("OnLeave",HideGTT);
	b:HookScript("OnClick",Button_OnClick);
	b:SetAttribute("type*","item");

	b.icon = b:CreateTexture(nil,"ARTWORK");
	b.icon:SetAllPoints();

	b.count = b:CreateFontString(nil,"ARTWORK","NumberFontNormal");
	b.count:SetPoint("BOTTOMRIGHT",b.icon,-3,3);

	b.cooldown = CreateFrame("Cooldown",nil,b,"CooldownFrameTemplate");
	b.cooldown:SetAllPoints();

	b.bind = b:CreateFontString(nil,"ARTWORK","NumberFontNormalSmallGray");
	b.bind:SetPoint("TOPLEFT",b.icon,3,-3);
	b.bind:SetPoint("TOPRIGHT",b.icon,-3,-3);
	b.bind:SetJustifyH("RIGHT");

	if (#core.items == 0) then
		b:SetPoint("TOPLEFT",core,12,-12);
	end

	core.items[#core.items + 1] = b;
	return b;
end

-- Add Button
local function AddButton(index,bag,slot,link,count)
	local btn = core.items[index] or CreateItemButton();

	btn.icon:SetTexture(select(10,GetItemInfo(link)));
	btn.count:SetText(count and count > 1 and count or "");

	btn.link = link;
	btn.name = GetItemInfo(link);

	btn:SetAttribute("bag",bag);
	btn:SetAttribute("slot",slot);

	if (index > 1) then
		btn:ClearAllPoints();
		if (cfg.vertical) then
			if (cfg.mirrored) then
				btn:SetPoint("BOTTOM",core.items[index - 1],"TOP",0,cfg.padding);
			else
				btn:SetPoint("TOP",core.items[index - 1],"BOTTOM",0,cfg.padding * -1);
			end
		else
			if (cfg.mirrored) then
				btn:SetPoint("RIGHT",core.items[index - 1],"LEFT",cfg.padding * -1,0);
			else
				btn:SetPoint("LEFT",core.items[index - 1],"RIGHT",cfg.padding,0);
			end
		end
	end

	btn:Show();
end

-- Check Item -- Az: Some items which starts a quest, are not marked as "Quest" in itemType or itemSubType. Ex: item:17008
local function CheckItem(link)
	local itemName, _, _, _, _, itemType, itemSubType = GetItemInfo(link);
	-- Check if ignored
	if (cfg.ignoreList) and (cfg.ignoreList[itemName]) then
		return;
	end
	-- Include predefinded items
	local itemId = tonumber(link:match("item:(%d+)"));
	for _, id in ipairs(qItems) do
		if (itemId == id) then
			return 1;
		end
	end
	-- Scan Tip -- Az: any reason we cant just check for more or equal to 4 lines, or would some quest items fail that check?
	core.tip:ClearLines();
	core.tip:SetHyperlink(link);
	local numLines = core.tip:NumLines();
	local line2 = (_G[modName.."TipTextLeft2"]:GetText() or "");
	if (numLines >= 3) and (itemType == "Quest" or itemSubType == "Quest" or line2 == ITEM_BIND_QUEST or line2 == GetZoneText()) or (numLines >= 4 and _G[modName.."TipTextLeft4"]:GetText() == ITEM_STARTS_QUEST) then
		for i = 3, numLines do
			local text = _G[modName.."TipTextLeft"..i]:GetText() or "";
			if (text:find("^"..ITEM_SPELL_TRIGGER_ONUSE) or text:find("^"..ITEM_STARTS_QUEST)) then
				return 1;
			end
		end
	end
end

--------------------------------------------------------------------------------------------------------
--                                               Update                                               --
--------------------------------------------------------------------------------------------------------

-- Request a Button Update
function core.RequestUpdate()
	updateTime = 0;
	core:SetScript("OnUpdate",OnUpdate);
end

-- Update Buttons
function core.UpdateButtons()
	-- Check if we are locked by combat
	if (InCombatLockdown()) then
		return;
	end
	-- locals
	local index = 1;
	-- Inventory
	for bag = 0, NUM_BAG_FRAMES do
		for slot = 1, GetContainerNumSlots(bag) do
			local link = GetContainerItemLink(bag,slot);
			if (link) and (CheckItem(link)) then
				AddButton(index,bag,slot,link,select(2,GetContainerItemInfo(bag,slot)));
				index = (index + 1);
			end
		end
	end
	-- Equipped Items
	for _, slotName in ipairs(slots) do
		local slotId = GetInventorySlotInfo(slotName);
		local link = GetInventoryItemLink("player",slotId);
		if (link) and (CheckItem(link)) then
			AddButton(index,nil,slotId,link);
			index = (index + 1);
		end
	end
	-- Set Shown Items
	shownItems = (index - 1);
	for i = index, #core.items do
		core.items[i]:Hide();
	end
	-- Update Misc
	core.UpdateBinding(1);
	core.UpdateCooldowns();
end

-- Update Cooldowns
function core.UpdateCooldowns()
	for i = 1, shownItems do
		local bag, slot = core.items[i]:GetAttribute("bag"), core.items[i]:GetAttribute("slot");
		if (bag) then
			CooldownFrame_SetTimer(core.items[i].cooldown,GetContainerItemCooldown(bag,slot));
		else
			CooldownFrame_SetTimer(core.items[i].cooldown,GetInventoryItemCooldown("player",slot));
		end
	end
end

-- Update Bind Text
function core.UpdateBinding(reBind)
	for i = 1, shownItems do
		if (not InCombatLockdown() and reBind and cfg.bindKey) and (shownItems == 1 or lastName == core.items[i].name) then
			SetBindingClick(cfg.bindKey,modName..i);
		end
		core.items[i].bind:SetText(GetBindingText(GetBindingKey("CLICK "..modName..i..":LeftButton"),"",1));
	end
end

--------------------------------------------------------------------------------------------------------
--                                               Events                                               --
--------------------------------------------------------------------------------------------------------

core:SetScript("OnEvent",function(self,event,...) self[event](self,event,...); end);
core:RegisterEvent("VARIABLES_LOADED");

-- Variables Loaded
function core:VARIABLES_LOADED(event)
	if (not QBar_Config) then
		QBar_Config = {};
	end
	cfg = setmetatable(QBar_Config,{ __index = defConfig });
	core:SetWidth(cfg.btnSize + 24);
	core:SetHeight(cfg.btnSize + 24);
	if (cfg.left and cfg.bottom) then
		core:SetPoint("BOTTOMLEFT",cfg.left,cfg.bottom);
	else
		core:SetPoint("CENTER");
	end
	SetEnabledStatus();
	core:SetScale(cfg.scale);
	-- Clear
	self:UnregisterEvent(event);
	self[event] = nil;
end

-- Update Cooldowns
function core:ACTIONBAR_UPDATE_COOLDOWN(event)
	if (shownItems > 0) then
		core.UpdateCooldowns();
	end
end

-- Bindings Update
function core:UPDATE_BINDINGS(event)
	core.UpdateBinding();
end

-- Inventory Changed
function core:UNIT_INVENTORY_CHANGED(event,unit)
	if (unit == "player") then
		core.RequestUpdate();
	end
end

-- Other Events
function core:PLAYER_REGEN_ENABLED(event)
	core.RequestUpdate();
end
core.BAG_UPDATE = core.PLAYER_REGEN_ENABLED;
core.ZONE_CHANGED_NEW_AREA = core.PLAYER_REGEN_ENABLED;

--------------------------------------------------------------------------------------------------------
--                                            Command Line                                            --
--------------------------------------------------------------------------------------------------------
_G["SLASH_"..modName.."1"] = "/qb";
SlashCmdList[modName] = function(cmd)
	-- Extract Parameters
	local param1, param2 = cmd:match("^([^%s]+)%s*(.*)$");
	param1 = (param1 and param1:lower() or cmd:lower());
	-- Enabled
	if (param1 == "toggle") then
		if (InCombatLockdown()) then
			AzMsg("|2"..modName.."|r Cannot toggle in combat");
		else
			cfg.enabled = not cfg.enabled;
			SetEnabledStatus();
			AzMsg("|2"..modName.."|r has now been |1"..(cfg.enabled and "enabled" or "disabled").."|r.");
		end
	-- Scale
	elseif (param1 == "scale") then
		if (param2 ~= "") and (type(tonumber(param2)) == "number") then
			cfg.scale = tonumber(param2);
			core:SetScale(cfg.scale);
			AzMsg("Button Scale set to |1"..cfg.scale.."|r.");
		end
	-- Size
	elseif (param1 == "size") then
		if (param2 ~= "") and (type(tonumber(param2)) == "number") then
			cfg.btnSize = tonumber(param2);
			for _, btn in ipairs(core.items) do
				btn:SetWidth(cfg.btnSize);
				btn:SetHeight(cfg.btnSize);
			end
			core:SetWidth(cfg.btnSize + 24);
			core:SetHeight(cfg.btnSize + 24);
			AzMsg("Button Size set to |1"..cfg.btnSize.."|r.");
		end
	-- Padding
	elseif (param1 == "padding") then
		if (param2 ~= "") and (type(tonumber(param2)) == "number") then
			cfg.padding = tonumber(param2);
			core.RequestUpdate();
			AzMsg("Button Padding set to |1"..cfg.padding.."|r.");
		end
	-- Tips
	elseif (param1 == "tips") then
		cfg.showTips = not cfg.showTips;
		AzMsg("Item tips |1"..(cfg.showTips and "enabled" or "disabled").."|r.");
	-- Vertical
	elseif (param1 == "vertical") then
		cfg.vertical = not cfg.vertical;
		core.RequestUpdate();
		AzMsg("Vertical alignment |1"..(cfg.vertical and "enabled" or "disabled").."|r.");
	-- Mirrored
	elseif (param1 == "mirror") then
		cfg.mirrored = not cfg.mirrored;
		core.RequestUpdate();
		AzMsg("Mirrored anchor |1"..(cfg.mirrored and "enabled" or "disabled").."|r.");
	-- Lock
	elseif (param1 == "lock") then
		SetLocked(core:IsMouseEnabled());
		AzMsg("QBar Button Frame is now |1"..(core:IsMouseEnabled() and "unlocked" or "locked").."|r.");
	-- Reset
	elseif (param1 == "reset") then
		core:ClearAllPoints();
		core:SetPoint("CENTER");
		cfg.left, cfg.bottom = core:GetLeft(), core:GetBottom();
	-- Bind
	elseif (param1 == "bind") then
		OpenBindingFrame();
	-- Clear Ignore
	elseif (param1 == "clearignore") then
		cfg.ignoreList = nil;
		core.RequestUpdate();
		AzMsg("Ignore List Cleared.");
	-- Invalid or No Command
	else
		UpdateAddOnMemoryUsage();
		AzMsg(format("----- |2%s|r |1%s|r ----- |1%.2core |2kb|r -----",modName,GetAddOnMetadata(modName,"Version"),GetAddOnMemoryUsage(modName)));
		AzMsg("The following |2parameters|r are valid for this addon:");
		AzMsg(" |2toggle|r = Toggles the mod being on or off");
		AzMsg(" |2scale 'value'|r = Sets the scale of the buttons");
		AzMsg(" |2size 'value'|r = Sets the unit size of the buttons");
		AzMsg(" |2padding 'value'|r = Space padding between the buttons");
		AzMsg(" |2tips|r = Toggles the showing of item tips when you mouse over buttons");
		AzMsg(" |2vertical|r = Sets whether or not the buttons anchor vertical or horizontal");
		AzMsg(" |2mirror|r = Mirror the anchor direction, top/bottom or left/right depending on the vertical setting");
		AzMsg(" |2lock|r = Toggles the button frame being locked, use this command to move the buttons around");
		AzMsg(" |2reset|r = Resets the position of QBar in case it has gotten off screen");
		AzMsg(" |2clearignore|r = Show all items again, by clearing the ignore list");
		AzMsg(" |2bind|r = Set a key binding for the last item used");
	end
end